#include <sourcemod>
#define PL_VERSION "1.2"
#pragma semicolon 1
new Handle:g_kValue = INVALID_HANDLE;
new Handle:db = INVALID_HANDLE;
new bool:colour = true;
new bool:roundend = false;
new bool:sqlite = false;
new bool:notify[MAXPLAYERS+1] = {true, ...};
new rankcount;
new rank[MAXPLAYERS+1];
new rating[MAXPLAYERS+1];
new score[MAXPLAYERS+1];
new enemyscore[MAXPLAYERS+1];
new deaths[MAXPLAYERS+1];
new sessionrating[MAXPLAYERS+1];
new sessionscore[MAXPLAYERS+1];
new sessionenemyscore[MAXPLAYERS+1];
new sessiondeaths[MAXPLAYERS+1];

new TEAM_MI6 = 2;
new TEAM_JANUS = 3;

public Plugin:myinfo =
{
	name = "Team ELO Ranking",
	author = "Mike + R_Hehl + djoslin",
	description = "Ranks players based on team score using the ELO rating system.",
	version = PL_VERSION,
	url = "http://www.fragtastic.org.uk/"
};

public OnPluginStart() {
	decl String:error[256];
	if(SQL_CheckConfig("elo")) {
		db = SQL_Connect("elo", true, error, sizeof(error));
	} else {
		db = SQL_Connect("storage-local", true, error, sizeof(error));
	}
	if(db==INVALID_HANDLE) {
		LogError("Could not connect to database: %s", error);
		return;
	}
	decl String:ident[16];
	SQL_ReadDriver(db, ident, sizeof(ident));
	if(strcmp(ident, "mysql", false)==0) {
		sqlite = false;
	} else if(strcmp(ident, "sqlite", false)==0) {
		sqlite = true;
	} else {
		LogError("Invalid database.");
		return;
	}

	//SQL_TQuery(db, SQLErrorCheckCallback, "DROP TABLE elostats");

	if(sqlite) {
		SQL_TQuery(db, SQLErrorCheckCallback, "CREATE TABLE IF NOT EXISTS elostats (steamid TEXT, name TEXT, rating INTEGER, score INTEGER, enemyscore INTEGER, deaths INTEGER, notify INTEGER)");
	} else {
		SQL_TQuery(db, SQLErrorCheckCallback, "CREATE TABLE IF NOT EXISTS elostats (steamid varchar(32) NOT NULL, name varchar(64) NOT NULL, rating int(8) NOT NULL, score int(8) NOT NULL, enemyscore int(8) NOT NULL, deaths int(8) NOT NULL, notify int(2) NOT NULL)");
	}
	CreateConVar("sm_elo_version", PL_VERSION, "ELO Ranking version.", FCVAR_PLUGIN|FCVAR_SPONLY|FCVAR_REPLICATED|FCVAR_NOTIFY);
	g_kValue = CreateConVar("sm_elo_k", "16", "K-Value for ELO Ranking.", FCVAR_PLUGIN|FCVAR_NOTIFY);
	RegConsoleCmd("sm_elo", Command_show);
	RegConsoleCmd("say", Command_say);
	RegConsoleCmd("say_team", Command_say);
	HookEvent("player_changename", Event_player_changename);
	HookEvent("player_death", Event_player_death);
	HookEvent("gameplay_event", Event_gameplay_event);
	decl String:mod[16];
	GetGameFolderName(mod, sizeof(mod));
	if(strcmp(mod, "dod")==0) {
		HookEvent("dod_restart_round", Event_round_start);
		HookEvent("dod_round_start", Event_round_start);
		HookEvent("dod_round_win", Event_round_win);
	} else if(strcmp(mod, "hl2mp")==0) {
		colour = false;
	}
}

public OnClientPutInServer(client) {
	if(!IsFakeClient(client)) {
		decl String:clientid[32], String:query[256];
		new userid = GetClientUserId(client);
		GetClientAuthString(client, clientid, sizeof(clientid));
		Format(query, sizeof(query), "SELECT rating,score,enemyscore,deaths,notify FROM elostats WHERE steamid='%s'", clientid);
		SQL_TQuery(db, SQLQueryConnect, query, userid);
		sessionrating[client] = 0;
		sessionscore[client] = 0;
		sessionenemyscore[client] = 0;
		sessiondeaths[client] = 0;
	}
	else {
		rating[client] = 1600;
		score[client] = 0;
		enemyscore[client] = 0;
		deaths[client] = 0;
		notify[client] = false;
	}
}

public SQLQueryConnect(Handle:owner, Handle:hndl, const String:error[], any:data) {
	new client;
	if((client = GetClientOfUserId(data))==0) {
		return;
	}
	if(hndl==INVALID_HANDLE) {
		LogError("Query failed: %s", error);
	} else {
		decl String:query[512], String:clientname[64], String:clientid[32];
		GetClientName(client, clientname, sizeof(clientname));
		decl String:safe_clientname[140];
		SQL_EscapeString(db, clientname, safe_clientname, sizeof(safe_clientname));
		ReplaceString(clientname, sizeof(clientname), "'", "");
		GetClientAuthString(client, clientid, sizeof(clientid));
		if(!SQL_MoreRows(hndl)) {
			if(sqlite) {
				Format(query, sizeof(query), "INSERT INTO elostats VALUES('%s', '%s', 1600, 0, 0, 0, 0)", clientid, safe_clientname);
				SQL_TQuery(db, SQLErrorCheckCallback, query);
			} else {
				Format(query, sizeof(query), "INSERT INTO elostats (steamid, name, rating, score, enemyscore, deaths, notify) VALUES ('%s', '%s', 1600, 0, 0, 0)", clientid, safe_clientname);
				SQL_TQuery(db, SQLErrorCheckCallback, query);
			}
			rating[client] = 1600;
			score[client] = 0;
			enemyscore[client] = 0;
			deaths[client] = 0;
			notify[client] = false;
			CreateTimer(15.0, EloWelcome, client);
		} else {
			Format(query, sizeof(query), "UPDATE elostats SET name='%s' WHERE steamid='%s'", client, safe_clientname, clientid);
			SQL_TQuery(db, SQLErrorCheckCallback, query);
			CreateTimer(15.0, EloWelcomeBack, client);
		}
		while(SQL_FetchRow(hndl)) {
			rating[client] = SQL_FetchInt(hndl, 0);
			score[client] = SQL_FetchInt(hndl, 1);
			enemyscore[client] = SQL_FetchInt(hndl, 2);
			deaths[client] = SQL_FetchInt(hndl, 3);
			notify[client] = SQL_FetchInt(hndl, 4)==0?false:true;
		}
	}
}

public SQLQueryRank(Handle:owner, Handle:hndl, const String:error[], any:data) {
	new client;
	if((client = GetClientOfUserId(data))==0) {
		return;
	}
	if(hndl==INVALID_HANDLE) {
		LogError("Query failed: %s", error);
	} else {
		rank[client] = SQL_FetchInt(hndl, 0)+1;
	}
	SQL_TQuery(db, SQLQueryCount, "SELECT COUNT(*) FROM elostats", GetClientUserId(client));
}

public SQLQueryCount(Handle:owner, Handle:hndl, const String:error[], any:data) {
	new client;
	if((client = GetClientOfUserId(data))==0) {
		return;
	}
	rankcount = SQL_FetchInt(hndl, 0);
	decl String:rankstr[16];
	new Float:tgpeg = enemyscore[client]==0 ? float(score[client]) : float(score[client]/enemyscore[client]);
	if((rank[client]%100)>10 && (rank[client]%100)<14) {
		Format(rankstr, sizeof(rankstr), "%ith", rank[client]);
	} else {
		switch(rank[client]%10) {
			case 0: Format(rankstr, sizeof(rankstr), "%ith", rank[client]);
			case 1: Format(rankstr, sizeof(rankstr), "%ist", rank[client]);
			case 2: Format(rankstr, sizeof(rankstr), "%ind", rank[client]);
			case 3: Format(rankstr, sizeof(rankstr), "%ird", rank[client]);
			case 4: Format(rankstr, sizeof(rankstr), "%ith", rank[client]);
			case 5: Format(rankstr, sizeof(rankstr), "%ith", rank[client]);
			case 6: Format(rankstr, sizeof(rankstr), "%ith", rank[client]);
			case 7: Format(rankstr, sizeof(rankstr), "%ith", rank[client]);
			case 8: Format(rankstr, sizeof(rankstr), "%ith", rank[client]);
			case 9: Format(rankstr, sizeof(rankstr), "%ith", rank[client]);
		}
	}
	decl String:buffer[64];
	new Handle:panel = CreatePanel();
	SetPanelTitle(panel, "ELO Ranking stats:");
	Format(buffer, sizeof(buffer), " ");
	DrawPanelText(panel, buffer);
	Format(buffer, sizeof(buffer), "Rank: %s (of %i)", rankstr, rankcount);
	DrawPanelText(panel, buffer);
	Format(buffer, sizeof(buffer), "Rating: %i", rating[client]);
	DrawPanelText(panel, buffer);
	Format(buffer, sizeof(buffer), " ");
	DrawPanelText(panel, buffer);
	Format(buffer, sizeof(buffer), "Team Goals: %i", score[client]);
	DrawPanelText(panel, buffer);
	Format(buffer, sizeof(buffer), "Enemy Goals: %i", enemyscore[client]);
	DrawPanelText(panel, buffer);
	Format(buffer, sizeof(buffer), "Goals Per Enemy Goal: %.2f", tgpeg);
	DrawPanelText(panel, buffer);
	Format(buffer, sizeof(buffer), " ");
	DrawPanelText(panel, buffer);
	Format(buffer, sizeof(buffer), "Deaths: %i", deaths[client]);
	DrawPanelText(panel, buffer);
	Format(buffer, sizeof(buffer), " ");
	DrawPanelText(panel, buffer);
	DrawPanelItem(panel, "Close");
	SendPanelToClient(panel, client, PanelHandlerNothing, 15);
	CloseHandle(panel);
}

public SQLQueryTop10(Handle:owner, Handle:hndl, const String:error[], any:data) {
	new client;
	if((client = GetClientOfUserId(data))==0) {
		return;
	}
	if(hndl==INVALID_HANDLE) {
		LogError("Query failed: %s", error);
	} else {
		decl String:qname[64], String:qrating[8], String:buffer[68];
		new i = 0;
		new Handle:panel = CreatePanel();
		SetPanelTitle(panel, "Top 10 players:");
		while(SQL_FetchRow(hndl)) {
			SQL_FetchString(hndl, 0, qname, sizeof(qname));
			SQL_FetchString(hndl, 1, qrating, sizeof(qrating));
			Format(buffer, sizeof(buffer), "%s - %s", qname, qrating);
			DrawPanelText(panel, buffer);
			i++;
		}
		DrawPanelItem(panel, "Close");
		SendPanelToClient(panel, client, PanelHandlerNothing, 15);
		CloseHandle(panel);
	}
}

public SQLErrorCheckCallback(Handle:owner, Handle:hndl, const String:error[], any:data) {
	if(!StrEqual("", error)) {
		LogError("Query failed: %s", error);
	}
}

public Action:EloWelcome(Handle:timer, any:client) {
	if(IsClientInGame(client)) {
		PrintToChat(client, "This appears to be your first visit.");
		PrintToChat(client, "You may say ^h/elo_notify^w in chat to enable ranking notifications.");
	}
}

public Action:EloWelcomeBack(Handle:timer, any:client) {
	if(IsClientInGame(client)) {
		PrintToChat(client, "^hWelcome back!^w Ranking is enabled.");
		PrintToChat(client, "Say commands: ^arank, top10, session, ratings, elo_notify");
	}
}

public Action:Event_player_changename(Handle:event, const String:name[], bool:dontBroadcast) {
	decl String:clientname[64], String:clientid[32], String:query[512];
	new client = GetClientOfUserId(GetEventInt(event, "userid"));
	GetClientName(client, clientname, sizeof(clientname));
	decl String:safe_clientname[140];
	SQL_EscapeString(db, clientname, safe_clientname, sizeof(safe_clientname));
	GetClientAuthString(client, clientid, sizeof(clientid));
	Format(query, sizeof(query), "UPDATE elostats SET name='%s' WHERE steamid='%s'", client, safe_clientname, clientid);
	SQL_TQuery(db, SQLErrorCheckCallback, query);
}

public Action:Event_player_death(Handle:event, const String:name[], bool:dontBroadcast) {

	new i = GetClientOfUserId(GetEventInt(event, "userid"));

	if(roundend) { return; }
	if(!IsClientInGame(i)) { return; }

	deaths[i]++;
	sessiondeaths[i]++;

	if(IsFakeClient(i)) { return; }
	
	decl String:clientid[32], String:query[512];
	GetClientAuthString(i, clientid, sizeof(clientid));
	Format(query, sizeof(query), "UPDATE elostats SET deaths=%i WHERE steamid='%s'", deaths[i], clientid);
	SQL_TQuery(db, SQLErrorCheckCallback, query);

}

public void TeamScored(winning_team) {

	// do not rank if the round is over
	if(roundend) { return; }

	// translate team indices
	decl winning_team_index;
	decl losing_team_index;
	if(winning_team == TEAM_MI6) { winning_team_index = 0; losing_team_index = 1; }
	else if(winning_team == TEAM_JANUS) { winning_team_index = 1; losing_team_index = 0; }
	else { return; }

	decl Float:total_team_rank[2] = { 0.0, 0.0 };
	decl Float:total_team_members[2] = { 0.0, 0.0 };
	decl Float:top_team_rank[2] = { 0.0, 0.0 };
	decl Float:average_team_rank[2] = { 0.0, 0.0 };
	decl user_rank;

	// grab in team information
	for(new i=1;i<=GetMaxClients();i++) {
		if(IsClientInGame(i)) {
			user_rank = rating[i];

			if(GetClientTeam(i) == TEAM_MI6) {
				total_team_rank[0] += float(user_rank);
				total_team_members[0] += 1.0;
				if(float(user_rank) > top_team_rank[0]) { top_team_rank[0] = float(user_rank); }
			}
			else if(GetClientTeam(i) == TEAM_JANUS) {
				total_team_rank[1] += float(user_rank);
				total_team_members[1] += 1.0;
				if(float(user_rank) > top_team_rank[1]) { top_team_rank[1] = float(user_rank); }
			}
		}
	}

	// do not rank people if there is an empty team
	if(total_team_members[0] < 1) { return; }
	if(total_team_members[1] < 1) { return; }

	// do not rank people if there is a team of one, and he is below average
	if(total_team_members[0] < 2 && top_team_rank[0] < 1560) { return; }
	if(total_team_members[1] < 2 && top_team_rank[1] < 1560) { return; }

	// grab averages
	average_team_rank[0] = total_team_rank[0] / total_team_members[0];
	average_team_rank[1] = total_team_rank[1] / total_team_members[1];
	

	decl user_team;
	decl Float:prob;
	decl diff;
	decl Float:enemy_rating;
	decl additional_team_score;
	decl String:clientid[32], String:query[256];

	// adjust ranks and scores
	for(new i=1;i<=GetMaxClients();i++) {
		if(!IsClientInGame(i)) { continue; }

		// only consider people who are on team MI6 or JANUS
		user_team = GetClientTeam(i);
		if(user_team != TEAM_MI6 && user_team != TEAM_JANUS) { continue; }

		if(user_team == winning_team) {
			// this person is on the winning team, add to their score
			score[i]++;
			sessionscore[i]++;

			enemy_rating = average_team_rank[losing_team_index];
			additional_team_score = 1;
		}
		else {
			// this person is on the losing team, add to their enemyscore
			enemyscore[i]++;
			sessionenemyscore[i]++;

			enemy_rating = average_team_rank[winning_team_index];
			additional_team_score = 0;
		}

		// Elo math time!
		prob = 1/(Pow(10.0, (enemy_rating-float(rating[i]))/400)+1);
		diff = RoundFloat(GetConVarFloat(g_kValue)*(additional_team_score-prob));

		// add/subtract to their rating
		rating[i] = rating[i] + diff;
		sessionrating[i] = sessionrating[i] + diff;

		// if we shouldn't save this rating, get out while we still can
		if(IsFakeClient(i)) { continue; }
		if(rating[i] < 1) { rating[i] = 1; continue; }

		// grab client's name
		decl String:clientname[64];
		GetClientName(i, clientname, sizeof(clientname));
		decl String:safe_clientname[140];
		SQL_EscapeString(db, clientname, safe_clientname, sizeof(safe_clientname));

		// save the rating
		GetClientAuthString(i, clientid, sizeof(clientid));
		Format(query, sizeof(query), "UPDATE elostats SET rating=%i,score=%i,enemyscore=%i,name='%s' WHERE steamid='%s'", rating[i], score[i], enemyscore[i], safe_clientname, clientid);
		SQL_TQuery(db, SQLErrorCheckCallback, query);

		// notify the player
		if(notify[i]) {
			if(user_team == winning_team) {
				PrintToChat(i, "^hYour team scored!^w You gained ^h%i^w points. (%i)", diff, rating[i]);
			}
			else {
				PrintToChat(i, "^oThe enemy team scored!^w You lost ^o%i^w points. (%i)", RoundFloat(FloatAbs(float(diff))), rating[i]);
			}
		}
	}
}

public Action:Event_gameplay_event(Handle:event, const String:name[], bool:dontBroadcast) {
	decl String:event_name[64];
	GetEventString(event, "name", event_name, sizeof(event_name));

	decl team_who_scored;
	team_who_scored = 0;

	// check if a team scored
	if(StrEqual(event_name, "mi6_scored", false)) {
		team_who_scored = TEAM_MI6;
	}
	else if(StrEqual(event_name, "janus_scored", false)) {
		team_who_scored = TEAM_JANUS;
	}

	// if so, call the team scored function
	if(team_who_scored > 0) {
		Call_StartFunction(null, TeamScored);
		Call_PushCell(team_who_scored);
		Call_Finish();
	}
}

public Action:Event_round_start(Handle:event, const String:name[], bool:dontBroadcast) {
	roundend = false;
}

public Action:Event_round_win(Handle:event, const String:name[], bool:dontBroadcast) {
	roundend = true;
}

public Action:Command_show(client, args) {
	for(new i=1;i<=GetMaxClients();i++) {
		if(IsClientInGame(i)) {
			PrintToConsole(client, "%N: %i", i, rating[i]);
		}
	}
	return Plugin_Handled;
}

public Action:Command_say(client, args) {
	decl String:text[192];
	if(GetCmdArgString(text, sizeof(text))<1) {
		return Plugin_Continue;
	}

	new startidx;
	if(text[strlen(text)-1]=='"') {
		text[strlen(text)-1] = '\0';
		startidx = 1;
	}

	if(strcmp(text[startidx], "/elo_notify", false)==0 || strcmp(text[startidx], "elo_notify", false)==0) {
		decl String:clientid[32], String:query[512];
		GetClientAuthString(client, clientid, sizeof(clientid));
		if(notify[client]==false) {
			notify[client] = true;
			if(colour) {
				PrintToChat(client, "\x01Rank notifications enabled. Say \x04/elo_notify\x01 again to disable them.");
			} else {
				PrintToChat(client, "Rank notifications enabled. Say /elo_notify again to disable them.");
			}
			Format(query, sizeof(query), "UPDATE elostats SET notify=1 WHERE steamid='%s'", clientid);
			SQL_TQuery(db, SQLErrorCheckCallback, query);
		} else {
			notify[client] = false;
			if(colour) {
				PrintToChat(client, "\x01Rank notifications disabled. Say \x04/elo_notify\x01 again to enable them.");
			} else {
				PrintToChat(client, "Rank notifications enabled. Say /elo_notify again to disable them.");
			}
			Format(query, sizeof(query), "UPDATE elostats SET notify=0 WHERE steamid='%s'", clientid);
			SQL_TQuery(db, SQLErrorCheckCallback, query);
		}

		return Plugin_Handled;

	} else if(strcmp(text[startidx], "ratings", false)==0 || strcmp(text[startidx], "/ratings", false)==0) {
		decl String:buffer[64];
		new Handle:panel = CreatePanel();
		SetPanelTitle(panel, "Ratings:");
		
		Format(buffer, sizeof(buffer), " ");
		DrawPanelText(panel, buffer);

		for(new i=1;i<=GetMaxClients();i++) {
			if(IsClientInGame(i)) {
				decl String:clientname[64];
				GetClientName(i, clientname, sizeof(clientname));
				ReplaceString(clientname, sizeof(clientname), "'", "");
				Format(buffer, sizeof(buffer), "%s: %i", clientname, rating[i]);
				DrawPanelText(panel, buffer);
			}
		}

		Format(buffer, sizeof(buffer), " ");
		DrawPanelText(panel, buffer);

		DrawPanelItem(panel, "Close");
		SendPanelToClient(panel, client, PanelHandlerNothing, 15);
		CloseHandle(panel);

		return Plugin_Handled;

	} else if(strcmp(text[startidx], "rank", false)==0 || strcmp(text[startidx], "/rank", false)==0
			|| strcmp(text[startidx], "stats", false)==0 || strcmp(text[startidx], "/stats", false)==0
			|| strcmp(text[startidx], "session", false)==0 || strcmp(text[startidx], "/session", false)==0) {
		
		decl String:clientid[32];
		GetClientAuthString(client, clientid, sizeof(clientid));
		
		if(StrContains(text[startidx], "rank", false)!=-1 || StrContains(text[startidx], "stats", false)!=-1) {
			decl String:query[512];
			Format(query, sizeof(query), "SELECT COUNT(*) FROM elostats WHERE rating>%i", rating[client]);
			SQL_TQuery(db, SQLQueryRank, query, GetClientUserId(client));
			return Plugin_Handled;
		
		} else {
			decl String:buffer[64];
			new Handle:panel = CreatePanel();
			SetPanelTitle(panel, "Session stats:");
			Format(buffer, sizeof(buffer), " ");
			DrawPanelText(panel, buffer);
			Format(buffer, sizeof(buffer), "Rating: %i", sessionrating[client]);
			DrawPanelText(panel, buffer);
			Format(buffer, sizeof(buffer), " ");
			DrawPanelText(panel, buffer);
			Format(buffer, sizeof(buffer), "Team Goals: %i", sessionscore[client]);
			DrawPanelText(panel, buffer);
			Format(buffer, sizeof(buffer), "Enemy Goals: %i", sessionenemyscore[client]);
			DrawPanelText(panel, buffer);
			Format(buffer, sizeof(buffer), " ");
			DrawPanelText(panel, buffer);
			Format(buffer, sizeof(buffer), "Deaths: %i", sessiondeaths[client]);
			DrawPanelText(panel, buffer);
			Format(buffer, sizeof(buffer), " ");
			DrawPanelText(panel, buffer);
			DrawPanelItem(panel, "Close");
			SendPanelToClient(panel, client, PanelHandlerNothing, 15);
			CloseHandle(panel);
		}

		return Plugin_Handled;

	} else if(strcmp(text[startidx], "top10", false)==0 || strcmp(text[startidx], "/top10", false)==0) {
		SQL_TQuery(db, SQLQueryTop10, "SELECT name,rating FROM elostats ORDER BY rating DESC LIMIT 0,10", GetClientUserId(client));
		return Plugin_Handled;
	}
	
	return Plugin_Continue;	
}

public PanelHandlerNothing(Handle:menu, MenuAction:action, param1, param2) {
	// Do nothing
}